# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import structlog

from orchestrator.db import ProcessSubscriptionTable, ProcessTable, ProductTable, SubscriptionTable, db
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters.filters import generic_filter
from orchestrator.db.filters.generic_filters import (
    generic_bool_filter,
    generic_is_like_filter,
    generic_values_in_column_filter,
)

logger = structlog.get_logger(__name__)


def product_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable, ProductTable)
        .filter(ProductTable.name.ilike("%" + value + "%"))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def tag_filter(query: SearchQuery, value: str) -> SearchQuery:
    tags = value.split("-")
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable, ProductTable)
        .filter(ProductTable.tag.in_(tags))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def subscriptions_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .filter(SubscriptionTable.description.ilike("%" + value + "%"))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def subscription_id_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = db.session.query(ProcessSubscriptionTable).join(SubscriptionTable)
    process_subscriptions = generic_is_like_filter(SubscriptionTable.subscription_id)(process_subscriptions, value)
    process_subscriptions = process_subscriptions.subquery()
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def target_filter(query: SearchQuery, value: str) -> SearchQuery:
    targets = value.split("-")
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .filter(ProcessSubscriptionTable.workflow_target.in_(targets))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


PROCESS_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = {
    "pid": generic_is_like_filter(ProcessTable.pid),
    "istask": generic_bool_filter(ProcessTable.is_task),
    "assignee": generic_values_in_column_filter(ProcessTable.assignee),
    "status": generic_values_in_column_filter(ProcessTable.last_status),
    "workflow": generic_is_like_filter(ProcessTable.workflow),
    "creator": generic_is_like_filter(ProcessTable.created_by),
    "product": product_filter,
    "tag": tag_filter,
    "subscription": subscriptions_filter,
    "subscriptionId": subscription_id_filter,
    "target": target_filter,
}


filter_processes = generic_filter(PROCESS_FILTER_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
from typing import Optional, Union

import structlog
from fastapi import Query, WebSocket
from fastapi.param_functions import Depends
from fastapi.routing import APIRouter
from oauth2_lib.fastapi import OIDCUserModel
from redis.asyncio import Redis as AIORedis

from orchestrator.api.error_handling import raise_status
from orchestrator.db import EngineSettingsTable
from orchestrator.schemas import EngineSettingsBaseSchema, EngineSettingsSchema, GlobalStatusEnum, WorkerStatus
from orchestrator.security import oidc_user
from orchestrator.services import settings
from orchestrator.services.processes import SYSTEM_USER, ThreadPoolWorkerStatus
from orchestrator.settings import ExecutorType, app_settings
from orchestrator.utils.json import json_dumps
from orchestrator.utils.redis import delete_keys_matching_pattern
from orchestrator.websocket import WS_CHANNELS, websocket_manager

router = APIRouter()
logger = structlog.get_logger()

CACHE_FLUSH_OPTIONS: dict[str, str] = {
    "all": "All caches",
}


@router.delete("/cache/{name}")
async def clear_cache(name: str) -> Union[int, None]:
    cache: AIORedis = AIORedis.from_url(app_settings.CACHE_URI)
    if name not in CACHE_FLUSH_OPTIONS:
        raise_status(HTTPStatus.BAD_REQUEST, "Invalid cache name")

    key_name = "orchestrator:*" if name == "all" else f"orchestrator:{name}:*"
    return await delete_keys_matching_pattern(cache, key_name)


@router.get("/cache-names")
def get_cache_names() -> dict[str, str]:
    return CACHE_FLUSH_OPTIONS


@router.put("/status", response_model=EngineSettingsSchema)
async def set_global_status(
    body: EngineSettingsBaseSchema, user: Optional[OIDCUserModel] = Depends(oidc_user)
) -> EngineSettingsSchema:
    """
    Update the global status of the engine to a new state.

    Args:
        body: The GlobalStatus object

    Returns:
        The updated global status object

    """

    engine_settings = EngineSettingsTable.query.with_for_update().one()

    result = settings.marshall_processes(engine_settings, body.global_lock)
    if not result:
        raise_status(
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
            detail="Something went wrong while updating the database aborting, possible manual intervention required",
        )
    if app_settings.SLACK_ENGINE_SETTINGS_HOOK_ENABLED:
        user_name = user.user_name if user else SYSTEM_USER
        settings.post_update_to_slack(EngineSettingsSchema.from_orm(result), user_name)

    status_response = generate_engine_status_response(result)
    if websocket_manager.enabled:
        # send engine status to socket.
        await websocket_manager.broadcast_data(
            [WS_CHANNELS.ENGINE_SETTINGS],
            {"engine-status": generate_engine_status_response(result)},
        )

    return status_response


@router.get("/worker-status", response_model=WorkerStatus)
def get_worker_status() -> WorkerStatus:
    """
    Return data on job workers and queues.

    Returns:
    - The number of queued jobs
    - The number of workers
    - The number of running jobs
    - The number of successful and unsuccessful jobs
    """

    if app_settings.EXECUTOR == ExecutorType.WORKER:
        from orchestrator.services.tasks import CeleryJobWorkerStatus

        return CeleryJobWorkerStatus()
    else:
        return ThreadPoolWorkerStatus()


@router.get("/status", response_model=EngineSettingsSchema)
def get_global_status() -> EngineSettingsSchema:
    """
    Retrieve the global status object.

    Returns:
        The global status of the engine

    """
    engine_settings = EngineSettingsTable.query.one()
    return generate_engine_status_response(engine_settings)


if app_settings.ENABLE_WEBSOCKETS:

    @router.websocket("/ws-status/")
    async def websocket_get_global_status(websocket: WebSocket, token: str = Query(...)) -> None:
        error = await websocket_manager.authorize(websocket, token)

        await websocket.accept()
        if error:
            await websocket_manager.disconnect(websocket, reason=error)
            return

        engine_settings = EngineSettingsTable.query.one()

        await websocket.send_text(json_dumps({"engine-status": generate_engine_status_response(engine_settings)}))

        channel = WS_CHANNELS.ENGINE_SETTINGS
        await websocket_manager.connect(websocket, channel)


def generate_engine_status_response(
    engine_settings: EngineSettingsTable,
) -> EngineSettingsSchema:
    """
    Generate the correct engine status response.

    Args:
        engine_settings: Engine settings database object

    Returns:
        Engine StatusEnum

    """

    if engine_settings.global_lock and engine_settings.running_processes > 0:
        result = EngineSettingsSchema.from_orm(engine_settings)
        result.global_status = GlobalStatusEnum.PAUSING
        return result
    elif engine_settings.global_lock and engine_settings.running_processes == 0:
        result = EngineSettingsSchema.from_orm(engine_settings)
        result.global_status = GlobalStatusEnum.PAUSED
        return result

    result = EngineSettingsSchema.from_orm(engine_settings)
    result.global_status = GlobalStatusEnum.RUNNING
    return result



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements process related API endpoints."""

from fastapi.param_functions import Depends
from fastapi.routing import APIRouter

from orchestrator.api.api_v1.endpoints import (
    fixed_input,
    health,
    processes,
    product_blocks,
    products,
    resource_types,
    settings,
    subscription_customer_descriptions,
    subscriptions,
    translations,
    user,
    workflows,
)
from orchestrator.security import opa_security_default

api_router = APIRouter()
api_router.include_router(
    fixed_input.router,
    prefix="/fixed_inputs",
    tags=["Core", "Fixed Inputs"],
    dependencies=[Depends(opa_security_default)],
)
api_router.include_router(
    processes.router, prefix="/processes", tags=["Core", "Processes"], dependencies=[Depends(opa_security_default)]
)
api_router.include_router(processes.ws_router, prefix="/processes", tags=["Core", "Processes"])

api_router.include_router(
    product_blocks.router,
    prefix="/product_blocks",
    tags=["Core", "Product Blocks"],
    dependencies=[Depends(opa_security_default)],
)
api_router.include_router(
    products.router, prefix="/products", tags=["Core", "Product"], dependencies=[Depends(opa_security_default)]
)
api_router.include_router(
    resource_types.router,
    prefix="/resource_types",
    tags=["Core", "Resource Types"],
    dependencies=[Depends(opa_security_default)],
)
api_router.include_router(
    subscriptions.router,
    prefix="/subscriptions",
    tags=["Core", "Subscriptions"],
    dependencies=[Depends(opa_security_default)],
)
api_router.include_router(
    subscription_customer_descriptions.router,
    prefix="/subscription_customer_descriptions",
    tags=["Core", "Subscription Customer Descriptions"],
    dependencies=[Depends(opa_security_default)],
)
api_router.include_router(
    user.router, prefix="/user", tags=["Core", "User"], dependencies=[Depends(opa_security_default)]
)
api_router.include_router(
    workflows.router, prefix="/workflows", tags=["Core", "Workflows"], dependencies=[Depends(opa_security_default)]
)
api_router.include_router(
    settings.router, prefix="/settings", tags=["Core", "Settings"], dependencies=[Depends(opa_security_default)]
)
api_router.include_router(health.router, prefix="/health", tags=["Core"])
api_router.include_router(
    translations.router,
    prefix="/translations",
    tags=["Core", "Translations"],
)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import secrets
import string
from pathlib import Path
from typing import List, Optional

from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from pydantic import BaseSettings, PostgresDsn, RedisDsn

from orchestrator.types import strEnum


class ExecutorType(strEnum):
    WORKER = "celery"
    THREADPOOL = "threadpool"


class AppSettings(BaseSettings):
    TESTING: bool = True
    SESSION_SECRET: str = "".join(secrets.choice(string.ascii_letters) for i in range(16))  # noqa: S311
    CORS_ORIGINS: str = "*"
    CORS_ALLOW_METHODS: List[str] = ["GET", "PUT", "PATCH", "POST", "DELETE", "OPTIONS", "HEAD"]
    CORS_ALLOW_HEADERS: List[str] = ["If-None-Match", "Authorization", "If-Match", "Content-Type"]
    CORS_EXPOSE_HEADERS: List[str] = [
        "Cache-Control",
        "Content-Language",
        "Content-Length",
        "Content-Type",
        "Expires",
        "Last-Modified",
        "Pragma",
        "Content-Range",
        "ETag",
    ]
    ENVIRONMENT: str = "local"
    EXECUTOR: str = ExecutorType.THREADPOOL
    WORKFLOWS_SWAGGER_HOST: str = "localhost"
    WORKFLOWS_GUI_URI: str = "http://localhost:3000"
    DATABASE_URI: PostgresDsn = "postgresql://nwa:nwa@localhost/orchestrator-core"  # type: ignore
    MAX_WORKERS: int = 5
    MAIL_SERVER: str = "localhost"
    MAIL_PORT: int = 25
    MAIL_STARTTLS: bool = False
    CACHE_URI: RedisDsn = "redis://localhost:6379/0"  # type: ignore
    CACHE_DOMAIN_MODELS: bool = False
    CACHE_HMAC_SECRET: Optional[str] = None  # HMAC signing key, used when pickling results in the cache
    ENABLE_DISTLOCK_MANAGER: bool = True
    DISTLOCK_BACKEND: str = "memory"
    CC_NOC: int = 0
    SERVICE_NAME: str = "orchestrator-core"
    LOGGING_HOST: str = "localhost"
    LOG_LEVEL: str = "DEBUG"
    SLACK_ENGINE_SETTINGS_HOOK_ENABLED: bool = False
    SLACK_ENGINE_SETTINGS_HOOK_URL: str = ""
    TRACING_ENABLED: bool = False
    TRACE_HOST: str = "http://localhost:4317"
    TRANSLATIONS_DIR: Optional[Path] = None
    WEBSOCKET_BROADCASTER_URL: str = "memory://"
    ENABLE_WEBSOCKETS: bool = True
    DISABLE_INSYNC_CHECK: bool = False
    DEFAULT_PRODUCT_WORKFLOWS: List[str] = ["modify_note"]
    SKIP_MODEL_FOR_MIGRATION_DB_DIFF: List[str] = []
    SERVE_GRAPHQL_UI: bool = True
    FEDEREATION_ENABLED: bool = False


class Oauth2Settings(BaseSettings):
    OAUTH2_ACTIVE: bool = False
    OAUTH2_RESOURCE_SERVER_ID: str = ""
    OAUTH2_RESOURCE_SERVER_SECRET: str = ""
    OAUTH2_TOKEN_URL: str = ""
    OIDC_CONF_WELL_KNOWN_URL: str = ""
    OPA_URL: str = "http://127.0.0.1:8181/v1/data/automation/authorization/allow"


app_settings = AppSettings()
oauth2_settings = Oauth2Settings()

# Tracer settings
tracer_provider = TracerProvider(resource=Resource.create({SERVICE_NAME: app_settings.SERVICE_NAME}))

otlp_exporter = OTLPSpanExporter(endpoint=app_settings.TRACE_HOST)
tracer_provider.add_span_processor(BatchSpanProcessor(otlp_exporter))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions for automatically generating docs for workflows and domain models.

WARNING: When developing on these auto generated docs,
    sphinx does not rebuild the files where these functions are called. To see your changes reflected you need to
    change or ``touch`` the file in question. Or use ``make clean docs``

WARNING2: Since we use the sphinx napoleon extension and the docstrings we generate are parsed by that.
    You cannot use all rst but you need to conform to google style docstring syntax. This also means that
    For any new header you probably need to add it to the docs config

    Also watch out for indentation. The resulting string needs to have the correct indentation.
"""

import inspect
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, get_origin
from uuid import UUID

from more_itertools import first

from orchestrator.types import SubscriptionLifecycle

if TYPE_CHECKING:
    from orchestrator.domain.base import ProductBlockModel, SubscriptionModel
    from orchestrator.workflow import Step, Workflow
    from orchestrator.workflows import LazyWorkflowInstance


INDENT1 = " " * 4
INDENT2 = " " * 8
INDENT3 = " " * 12


def make_ref(field_type: Any, shorten: bool = True) -> str:
    if hasattr(field_type, "__qualname__"):
        ref = f"{'~' if shorten else ''}{field_type.__module__}.{field_type.__qualname__.split('.<locals>')[0].split('[')[0]}"
    elif get_origin(field_type):
        ref = make_ref(get_origin(field_type))
    else:
        ref = str(field_type)
    return ref


def get_doc_title(field_type: Any) -> str:
    return first((field_type.__doc__ or "").splitlines(), "")


def make_field_doc(field_name: str, field_type: Any) -> str:
    r"""Make docstring for domain model field.

    >>> make_field_doc("int_field", int)
    '        int_field:\n            Type :class:`~builtins.int`'

    >>> class Something:
    ...     "Doc title"
    >>> make_field_doc("int_field", Something)
    '        int_field:\n            Doc title\n\n            Type :class:`~orchestrator.utils.docs.Something`'
    """

    if field_type in (int, bool, str, UUID):
        type_doc_str = ""
    else:
        type_doc_str = f"{INDENT3}" + get_doc_title(field_type) + "\n\n"

    return f"{INDENT2}{field_name}:\n{type_doc_str}{INDENT3}Type :class:`{make_ref(field_type)}`"


def make_product_block_docstring(
    block: Type["ProductBlockModel"], lifecycle: Optional[List[SubscriptionLifecycle]] = None
) -> str:
    lifecycle_str = f"\n\nValid for statuses: {', '.join(lifecycle) if lifecycle else 'all others'}"

    if SubscriptionLifecycle.ACTIVE not in (lifecycle or []):
        return f"{lifecycle_str}\n\nSee `active` version."

    siv_strings = []
    for field_name, field_type in block._non_product_block_fields_.items():
        siv_strings.append(make_field_doc(field_name, field_type))

    product_block_strings = []
    for field_name, field_type in block._product_block_fields_.items():
        product_block_strings.append(make_field_doc(field_name, field_type))

    siv_string = ""
    if siv_strings:
        siv_string = "\nInstance Values:\n\n" + "\n".join(siv_strings)

    product_block_string = ""
    if product_block_strings:
        product_block_string = "\nBlocks:\n\n" + "\n".join(product_block_strings)

    return (block.__doc__ or "") + f"\n\n{lifecycle_str}\n\n{siv_string}{product_block_string}\n"


def make_subscription_model_docstring(
    model: Type["SubscriptionModel"], lifecycle: Optional[List[SubscriptionLifecycle]] = None
) -> str:
    lifecycle_str = f"\n\nValid for statuses: {', '.join(lifecycle) if lifecycle else 'all others'}"

    if SubscriptionLifecycle.ACTIVE not in (lifecycle or []):
        return f"{lifecycle_str}\n\nSee `active` version."

    fixed_input_strings = []
    for field_name, field_type in model._non_product_block_fields_.items():
        fixed_input_strings.append(make_field_doc(field_name, field_type))

    product_block_strings = []
    for field_name, field_type in model._product_block_fields_.items():
        product_block_strings.append(make_field_doc(field_name, field_type))

    fixed_input_string = ""
    if fixed_input_strings:
        fixed_input_string = "\nFixed Inputs:\n\n" + "\n".join(fixed_input_strings)

    product_block_string = ""
    if product_block_strings:
        product_block_string = "\nBlocks:\n\n" + "\n".join(product_block_strings)

    return (model.__doc__ or "") + f"\n\n{lifecycle_str}\n\n{fixed_input_string}{product_block_string}"


def make_workflow_doc(wf: "Workflow") -> str:
    def make_step_doc(step: "Step") -> str:
        doc = inspect.getdoc(step)
        doc_title = doc.splitlines()[0] if doc else ""

        return (
            f"{INDENT1}#. :func:`{step.__name__.strip('_')} <{make_ref(step, shorten=False)}>`\n\n{INDENT2}{doc_title}"
        )

    steps_string = "\n".join(make_step_doc(step) for step in wf.steps)

    return f"{(wf.__doc__ or '')}\n\nSteps:\n\n{steps_string}\n"


def make_workflow_index_doc(all_workflows: Dict[str, "LazyWorkflowInstance"]) -> str:
    workflow_list_str = "\n".join(map(lambda wf: f"* :obj:`~{wf[1].package}.{wf[0]}`", all_workflows.items()))
    return f"\nWorkflows\n---------\n\n{workflow_list_str}\n"


def make_product_type_index_doc(subscription_model_registry: Dict[str, Type["SubscriptionModel"]]) -> str:
    product_types_str = "\n".join(
        map(
            lambda product: f"* :class:`~{product.__module__}.{product.__qualname__}`",
            set(subscription_model_registry.values()),
        )
    )

    return f"\nProduct Types\n---------\n\n{product_types_str}\n"



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
from http import HTTPStatus
from typing import Union

from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.api.test_subscriptions import (  # noqa: F401
    PORT_A_SUBSCRIPTION_ID,
    SERVICE_SUBSCRIPTION_ID,
    SSP_SUBSCRIPTION_ID,
    seed,
)

subscription_fields = [
    "endDate",
    "description",
    "subscriptionId",
    "startDate",
    "productId",
    "status",
    "insync",
    "note",
    "product",
    "productBlocks",
]
subscription_product_fields = [
    "productId",
    "name",
    "description",
    "productType",
    "status",
    "tag",
    "createdAt",
    "endDate",
]


def get_subscriptions_query(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query SubscriptionQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!]) {
  subscriptions(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy) {
    page {
      description
      subscriptionId
      productId
      status
      insync
      note
      startDate
      endDate
      productBlocks {
        id
        parent
        ownerSubscriptionId
        resourceTypes
      }
      product {
        productId
        name
        description
        productType
        status
        tag
        createdAt
        endDate
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def get_subscriptions_query_with_relations(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query SubscriptionQuery($first: Int!, $after: Int!, $sortBy: [GraphqlSort!], $filterBy: [GraphqlFilter!]) {
  subscriptions(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy) {
    page {
      description
      subscriptionId
      productId
      status
      insync
      note
      startDate
      endDate
      processes {
        page {
          assignee
          createdBy
          failedReason
          isTask
          lastStep
          traceback
          id
          lastModified
          started
          workflowName
          status
          step
          product
        }
      }
      dependsOnSubscriptions {
        page {
          description
          subscriptionId
          productId
          status
          insync
          note
          startDate
          endDate
        }
      }
      inUseBySubscriptions {
        page {
          description
          subscriptionId
          productId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def test_subscriptions_single_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(7)
    data = get_subscriptions_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 6,
        "totalItems": "7",
    }

    for subscription in subscriptions:
        for field in subscription_fields:
            assert field in subscription
        for field in subscription_product_fields:
            assert field in subscription["product"]


def test_subscriptions_has_next_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "30",
    }

    for subscription in subscriptions:
        for field in subscription_fields:
            assert field in subscription


def test_subscriptions_has_previous_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(after=1)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 10

    assert pageinfo == {
        "hasPreviousPage": True,
        "hasNextPage": True,
        "startCursor": 1,
        "endCursor": 10,
        "totalItems": "30",
    }


def test_subscriptions_sorting_asc(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "startDate", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "30",
    }

    for i in range(0, 8):
        assert subscriptions[i]["startDate"] < subscriptions[i + 1]["startDate"]


def test_subscriptions_sorting_desc(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "startDate", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "30",
    }

    for i in range(0, 8):
        assert subscriptions[i + 1]["startDate"] < subscriptions[i]["startDate"]


def test_subscriptions_sorting_product_tag_asc(test_client, generic_subscription_1, generic_subscription_2):
    # when

    data = get_subscriptions_query(sort_by=[{"field": "tag", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": "2",
    }

    product_tag_list = [subscription["product"]["tag"] for subscription in subscriptions]
    assert product_tag_list == ["GEN1", "GEN2"]


def test_subscriptions_sorting_product_tag_desc(test_client, generic_subscription_1, generic_subscription_2):
    # when

    data = get_subscriptions_query(sort_by=[{"field": "tag", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": "2",
    }

    product_tag_list = [subscription["product"]["tag"] for subscription in subscriptions]
    assert product_tag_list == ["GEN2", "GEN1"]


def test_subscriptions_sorting_invalid_field(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "start_date", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": "30",
    }

    assert "errors" in result
    assert result["errors"] == [
        {
            "message": "Invalid sort arguments",
            "path": [None, "subscriptions", "Query"],
            "extensions": {
                "invalid_sorting": [{"field": "start_date", "order": "DESC"}],
                "valid_sort_keys": [
                    "productId",
                    "name",
                    "description",
                    "productType",
                    "tag",
                    "status",
                    "createdAt",
                    "endDate",
                    "subscriptionId",
                    "customerId",
                    "insync",
                    "startDate",
                    "note",
                ],
            },
        }
    ]


def test_subscriptions_sorting_invalid_order(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "start_date", "order": "test"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    assert not result["data"]
    assert "errors" in result
    assert "Value 'test' does not exist in 'SortOrder'" in result["errors"][0]["message"]


def test_subscriptions_filtering_on_status(test_client, product_type_1_subscriptions_factory, generic_product_type_1):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)

    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    subscription_1 = GenericProductOne.from_subscription(subscription_ids[0])
    subscription_1 = subscription_1.from_other_lifecycle(subscription_1, SubscriptionLifecycle.TERMINATED)
    subscription_1.save()
    subscription_9 = GenericProductOne.from_subscription(subscription_ids[8])
    subscription_9 = subscription_9.from_other_lifecycle(subscription_9, SubscriptionLifecycle.TERMINATED)
    subscription_9.save()

    data = get_subscriptions_query(filter_by=[{"field": "status", "value": SubscriptionLifecycle.TERMINATED}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": "2",
    }

    assert subscriptions[0]["subscriptionId"] == str(subscription_1.subscription_id)
    assert subscriptions[0]["status"] == SubscriptionLifecycle.TERMINATED
    assert subscriptions[1]["subscriptionId"] == str(subscription_9.subscription_id)
    assert subscriptions[1]["status"] == SubscriptionLifecycle.TERMINATED


def test_subscriptions_range_filtering_on_start_date(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)

    data = get_subscriptions_query(first=1)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    first_subscription = response.json()["data"]["subscriptions"]["page"][0]

    first_subscription_date = datetime.datetime.fromisoformat(first_subscription["startDate"])
    higher_then_date = first_subscription_date.isoformat()
    lower_then_date = (first_subscription_date + datetime.timedelta(days=3)).isoformat()

    data = get_subscriptions_query(
        filter_by=[
            {"field": "startDateGte", "value": higher_then_date},
            {"field": "startDateLte", "value": lower_then_date},
        ]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": "4",
    }

    for subscription in subscriptions:
        assert higher_then_date <= subscription["startDate"] <= lower_then_date


def test_subscriptions_filtering_with_invalid_filter(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1
):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)

    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    subscription_1 = GenericProductOne.from_subscription(subscription_ids[0])
    subscription_1 = subscription_1.from_other_lifecycle(subscription_1, SubscriptionLifecycle.TERMINATED)
    subscription_1.save()
    subscription_9 = GenericProductOne.from_subscription(subscription_ids[8])
    subscription_9 = subscription_9.from_other_lifecycle(subscription_9, SubscriptionLifecycle.TERMINATED)
    subscription_9.save()

    data = get_subscriptions_query(
        filter_by=[
            {"field": "status", "value": SubscriptionLifecycle.TERMINATED},
            {"field": "test", "value": "invalid"},
        ]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    errors = result["errors"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert errors == [
        {
            "message": "Invalid filter arguments",
            "path": [None, "subscriptions", "Query"],
            "extensions": {
                "invalid_filters": [{"field": "test", "value": "invalid"}],
                "valid_filter_keys": [
                    "subscriptionId",
                    "subscriptionIds",
                    "description",
                    "status",
                    "product",
                    "insync",
                    "note",
                    "statuses",
                    "tags",
                    "tag",
                    "tsv",
                    "startDate",
                    "endDate",
                    "startDateGt",
                    "startDateGte",
                    "startDateLt",
                    "startDateLte",
                    "startDateNe",
                    "endDateGt",
                    "endDateGte",
                    "endDateLt",
                    "endDateLte",
                    "endDateNe",
                ],
            },
        }
    ]
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": "2",
    }

    for subscription in subscriptions:
        assert subscription["status"] == SubscriptionLifecycle.TERMINATED


def test_single_subscription(test_client, product_type_1_subscriptions_factory):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)
    subscription_id = subscription_ids[10]
    data = get_subscriptions_query(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["productBlocks"] == [
        {
            "id": 0,
            "ownerSubscriptionId": subscription_id,
            "parent": None,
            "resourceTypes": {"label": None, "name": "PB_1", "rt1": "Value1"},
        },
        {
            "id": 1,
            "ownerSubscriptionId": subscription_id,
            "parent": None,
            "resourceTypes": {"label": None, "name": "PB_2", "rt2": 42, "rt3": "Value2"},
        },
    ]


def test_single_subscription_with_processes(
    test_client,
    product_type_1_subscriptions_factory,
    mocked_processes,
    mocked_processes_resumeall,  # noqa: F811
    generic_subscription_2,  # noqa: F811
    generic_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = generic_subscription_1
    data = get_subscriptions_query_with_relations(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["processes"]["page"][0]["id"] == str(mocked_processes[0])


def test_single_subscription_with_depends_on_subscriptions(
    test_client, product_type_1_subscriptions_factory, seed  # noqa: F811
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = SERVICE_SUBSCRIPTION_ID
    data = get_subscriptions_query_with_relations(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["processes"]["page"]) == 0
    result_ids = [sub["subscriptionId"] for sub in subscriptions[0]["dependsOnSubscriptions"]["page"]]
    assert PORT_A_SUBSCRIPTION_ID in result_ids
    assert SSP_SUBSCRIPTION_ID in result_ids
    assert len(subscriptions[0]["inUseBySubscriptions"]["page"]) == 0


def test_single_subscription_with_in_use_by_subscriptions(
    test_client, product_type_1_subscriptions_factory, seed  # noqa: F811
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = PORT_A_SUBSCRIPTION_ID
    data = get_subscriptions_query_with_relations(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": "1",
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["processes"]["page"]) == 0
    assert len(subscriptions[0]["dependsOnSubscriptions"]["page"]) == 0
    assert subscriptions[0]["inUseBySubscriptions"]["page"][0]["subscriptionId"] == SERVICE_SUBSCRIPTION_ID



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import structlog

from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.db.filters import Filter
from orchestrator.db.filters.subscription import filter_subscriptions
from orchestrator.db.range import apply_range_to_query
from orchestrator.db.sorting import Sort, sort_subscriptions
from orchestrator.graphql.pagination import Connection, PageInfo
from orchestrator.graphql.schemas.subscription import SubscriptionType
from orchestrator.graphql.types import CustomInfo, GraphqlFilter, GraphqlSort
from orchestrator.graphql.utils.create_resolver_error_handler import create_resolver_error_handler

logger = structlog.get_logger(__name__)


async def resolve_subscriptions(
    info: CustomInfo,
    filter_by: Union[list[GraphqlFilter], None] = None,
    sort_by: Union[list[GraphqlSort], None] = None,
    first: int = 10,
    after: int = 0,
) -> Connection[SubscriptionType]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.info("resolve_subscription() called", range=[after, after + first], sort=sort_by, filter=pydantic_filter_by)

    query = SubscriptionTable.query.join(ProductTable)

    query = filter_subscriptions(query, pydantic_filter_by, _error_handler)
    query = sort_subscriptions(query, pydantic_sort_by, _error_handler)
    total = query.count()
    query = apply_range_to_query(query, after, first)

    subscriptions = query.all()
    has_next_page = len(subscriptions) > first

    # exclude last item as it was fetched to know if there is a next page
    subscriptions = subscriptions[:first]
    subscriptions_length = len(subscriptions)
    start_cursor = after if subscriptions_length else None
    end_cursor = after + subscriptions_length - 1
    page_subscriptions = [SubscriptionType.from_pydantic(p) for p in subscriptions]

    return Connection(
        page=page_subscriptions,
        page_info=PageInfo(
            has_previous_page=bool(after),
            has_next_page=has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            total_items=total if total else None,
        ),
    )



import json
from typing import List, Union

from orchestrator.cli.database import migrate_domain_models
from orchestrator.db import db
from orchestrator.db.models import ProductTable, ResourceTypeTable
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


def test_migrate_domain_models_new_product(test_product_type_one, test_product_sub_block_one_db):
    _, _, ProductTypeOneForTest = test_product_type_one
    inputs = {
        "TestProductOne": {
            "description": "test description",
            "tag": "test_tag",
            "product_type": "test_type",
            "test_fixed_input": "test value",
        },
        "ProductBlockOneForTest": {"description": "product block description", "tag": "test_block_tag"},
        "int_field": {"ProductBlockOneForTest": "1"},
        "str_field": {"ProductBlockOneForTest": "test"},
        "list_field": {"description": "list field desc", "ProductBlockOneForTest": "list test"},
    }
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=json.dumps(inputs))

    assert len(upgrade_sql) == 9
    assert len(downgrade_sql) == 18

    product_id = (
        ProductTable.query.where(ProductTable.name == "TestProductOne").with_entities(ProductTable.product_id).all()
    )
    assert not product_id

    temp_product = ProductTable(
        name="TestProductOne",
        description=inputs["TestProductOne"]["description"],
        product_type=inputs["TestProductOne"]["product_type"],
        tag=inputs["TestProductOne"]["tag"],
        status="active",
    )
    db.session.add(temp_product)
    db.session.commit()

    expected_old_diff = {
        "TestProductOne": {
            "missing_fixed_inputs_in_db": {"test_fixed_input"},
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {
                    "missing_product_blocks_in_db": {"SubBlockOneForTest"},
                    "missing_resource_types_in_db": {"int_field", "list_field", "str_field"},
                }
            },
            "missing_product_blocks_in_db": {"ProductBlockOneForTest"},
        }
    }

    diff = ProductTypeOneForTest.diff_product_in_database(temp_product.product_id)
    assert diff == expected_old_diff

    db.session.delete(temp_product)
    db.session.commit()

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    product_id = (
        ProductTable.query.where(ProductTable.name == "TestProductOne").with_entities(ProductTable.product_id).all()
    )
    assert product_id[0][0]
    diff = ProductTypeOneForTest.diff_product_in_database(product_id[0][0])
    assert diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    product_id = (
        ProductTable.query.where(ProductTable.name == "TestProductOne").with_entities(ProductTable.product_id).all()
    )
    assert not product_id


def test_migrate_domain_models_new_fixed_input(test_product_one, test_product_type_one, test_product_block_one):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        new_fixed_input: bool
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"TestProductOne": {"new_fixed_input": "test"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    expected_old_diff = {"TestProductOne": {"missing_fixed_inputs_in_db": {"new_fixed_input"}}}

    assert len(upgrade_sql) == 1
    assert len(downgrade_sql) == 1

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_rename_fixed_input(test_product_one, test_product_type_one, test_product_block_one):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        changed_fixed_input: bool
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    updates = json.dumps({"fixed_inputs": {"TestProductOne": {"test_fixed_input": "changed_fixed_input"}}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, updates=updates)

    expected_old_diff = {
        "TestProductOne": {
            "missing_fixed_inputs_in_db": {"changed_fixed_input"},
            "missing_fixed_inputs_in_model": {"test_fixed_input"},
        }
    }

    assert len(upgrade_sql) == 1
    assert len(downgrade_sql) == 1

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_new_product_block(test_product_one, test_product_type_one, test_product_block_one):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
        str_field: str

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        new_block: TestBlock
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps(
        {
            "test block": {"description": "test block description", "tag": "test_block_tag"},
            "str_field": {"test block": "test"},
        }
    )
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    expected_old_diff = {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {"test block": {"missing_resource_types_in_db": {"str_field"}}},
            "missing_product_blocks_in_db": {"test block"},
        }
    }

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 5

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_new_product_block_on_product_block(
    test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    _, _, SubBlockOneForTest = test_product_sub_block_one

    class TestBlock(ProductBlockModel, product_block_name="test block", lifecycle=[SubscriptionLifecycle.ACTIVE]):
        str_field: str

    class ProductBlockOneForTestUpdated(ProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: List[SubBlockOneForTest]
        str_field: str
        int_field: int
        list_field: List[int]
        new_block: TestBlock

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps(
        {
            "test block": {"description": "test block description", "tag": "test_block_tag"},
            "str_field": {"test block": "test"},
        }
    )
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    expected_old_diff = {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {"missing_product_blocks_in_db": {"test block"}},
                "test block": {"missing_resource_types_in_db": {"str_field"}},
            }
        }
    }

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 5

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_new_resource_type(
    test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestUpdated(ProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: List[SubBlockOneForTest]
        int_field: int
        str_field: str
        list_field: List[int]
        new_int_field: int

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"new_int_field": {"ProductBlockOneForTest": 1, "description": "test new int field type"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs)

    expected_old_diff = {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {"missing_resource_types_in_db": {"new_int_field"}}
            }
        }
    }

    assert len(upgrade_sql) == 2
    assert len(downgrade_sql) == 4

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_rename_resource_type(
    test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestUpdated(ProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: List[SubBlockOneForTest]
        str_field: str
        int_field: int
        new_list_field: List[int]

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    updates = json.dumps({"resource_types": {"list_field": "new_list_field"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, updates=updates)

    expected_old_diff = {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {
                    "missing_resource_types_in_db": {"new_list_field"},
                    "missing_resource_types_in_model": {"list_field"},
                }
            }
        }
    }

    assert len(upgrade_sql) == 1
    assert "UPDATE" in upgrade_sql[0]
    assert len(downgrade_sql) == 1
    assert "UPDATE" in downgrade_sql[0]

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_rename_and_relate_resource_type(
    test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class SubBlockOneForTestNewResource(SubBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        int_field: int
        str_field: str
        new_list_field: List[int]  # add/relate renamed resource type to block

    class ProductBlockOneForTestUpdated(ProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTestNewResource
        sub_block_2: SubBlockOneForTestNewResource
        sub_block_list: List[SubBlockOneForTestNewResource]
        str_field: str
        int_field: int
        new_list_field: List[int]  # renamed from 'list_field'

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"new_list_field": {"SubBlockOneForTest": "test"}})
    updates = json.dumps({"resource_types": {"list_field": "new_list_field"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs, updates)

    expected_old_diff = {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {
                    "missing_resource_types_in_db": {"new_list_field"},
                    "missing_resource_types_in_model": {"list_field"},
                },
                "SubBlockOneForTest": {"missing_resource_types_in_db": {"new_list_field"}},
            }
        }
    }

    assert len(upgrade_sql) == 2
    assert [sql_stmt for sql_stmt in upgrade_sql if "UPDATE" in sql_stmt]
    assert len(downgrade_sql) == 3
    assert [sql_stmt for sql_stmt in downgrade_sql if "UPDATE" in sql_stmt]

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    downgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert downgrade_diff == expected_old_diff


def test_migrate_domain_models_rename_and_relate_and_remove_resource_type(
    test_product_sub_list_union,
    test_product_type_sub_list_union,
    test_product_block_with_list_union,
    test_product_sub_block_one,
    test_product_sub_block_two,
):
    _, ProductSubListUnionProvisioning, _ = test_product_type_sub_list_union
    _, _, ProductBlockWithListUnionForTest = test_product_block_with_list_union
    _, _, SubBlockOneForTest = test_product_sub_block_one
    _, _, SubBlockTwoForTest = test_product_sub_block_two

    class SubBlockOneForTestChanged(SubBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        changed_int_field: int

    class SubBlockTwoForTestChanged(SubBlockTwoForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        int_field_2: int
        changed_int_field: int

    class ProductBlockWithListUnionForTestNew(
        ProductBlockWithListUnionForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        list_union_blocks: List[Union[SubBlockTwoForTestChanged, SubBlockOneForTestChanged]]
        changed_int_field: int
        str_field: str
        list_field: List[int]

    class ProductSubListUnionTest(ProductSubListUnionProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: ProductBlockWithListUnionForTestNew

    SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"] = ProductSubListUnionTest

    inputs = json.dumps({"changed_int_field": {"SubBlockOneForTest": "test", "SubBlockTwoForTest": "test"}})
    updates = json.dumps({"resource_types": {"int_field": "changed_int_field"}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs, updates)

    expected_old_diff = {
        "ProductSubListUnion": {
            "missing_in_depends_on_blocks": {
                "ProductBlockWithListUnionForTest": {
                    "missing_resource_types_in_db": {"changed_int_field"},
                    "missing_resource_types_in_model": {"int_field"},
                },
                "SubBlockOneForTest": {
                    "missing_resource_types_in_db": {"changed_int_field"},
                    "missing_resource_types_in_model": {"int_field", "str_field"},
                },
                "SubBlockTwoForTest": {"missing_resource_types_in_db": {"changed_int_field"}},
            }
        }
    }

    assert len(upgrade_sql) == 4
    assert [sql_stmt for sql_stmt in upgrade_sql if "UPDATE" in sql_stmt]
    assert len(downgrade_sql) == 3
    assert [sql_stmt for sql_stmt in downgrade_sql if "UPDATE" in sql_stmt]

    before_diff = ProductSubListUnionTest.diff_product_in_database(test_product_sub_list_union)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductSubListUnionTest.diff_product_in_database(test_product_sub_list_union)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()


def test_migrate_domain_models_update_block_resource_type(
    test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTestUpdated(ProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: List[SubBlockOneForTest]
        str_field: str
        new_int_field: int
        list_field: List[int]

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTestUpdated

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    inputs = json.dumps({"new_int_field": {"description": "test new int field type"}})
    updates = json.dumps({"block_resource_types": {"ProductBlockOneForTest": {"int_field": "new_int_field"}}})
    upgrade_sql, downgrade_sql = migrate_domain_models("example", True, inputs=inputs, updates=updates)

    expected_old_diff = {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {
                    "missing_resource_types_in_db": {"new_int_field"},
                    "missing_resource_types_in_model": {"int_field"},
                }
            }
        }
    }

    def test_expected_before_upgrade():
        before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
        assert before_diff == expected_old_diff
        int_field_resource = ResourceTypeTable.query.where(ResourceTypeTable.resource_type == "int_field").all()
        new_int_field_resource = ResourceTypeTable.query.where(ResourceTypeTable.resource_type == "new_int_field").all()
        assert int_field_resource
        assert not new_int_field_resource

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 4

    test_expected_before_upgrade()

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    upgrade_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert upgrade_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    test_expected_before_upgrade()


def test_migrate_domain_models_remove_product(test_product_one, test_product_type_one):
    _, _, ProductTypeOneForTest = test_product_type_one
    del SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"]

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    assert len(upgrade_sql) == 7
    assert len(downgrade_sql) == 0

    before_diff = ProductTypeOneForTest.diff_product_in_database(test_product_one)
    assert before_diff == {}

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    product_id = (
        ProductTable.query.where(ProductTable.name == "TestProductOne").with_entities(ProductTable.product_id).all()
    )
    assert not product_id

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTest


def test_migrate_domain_models_remove_fixed_input(test_product_one, test_product_type_one, test_product_block_one):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, _, ProductBlockOneForTest = test_product_block_one

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    expected_old_diff = {"TestProductOne": {"missing_fixed_inputs_in_model": {"test_fixed_input"}}}

    assert len(upgrade_sql) == 1
    assert len(downgrade_sql) == 1

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == expected_old_diff


def test_migrate_domain_models_remove_product_block(test_product_one, test_product_type_one):
    _, _, ProductTypeOneForTest = test_product_type_one

    class ProductTypeOneForTestNew(ProductTypeOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    assert len(upgrade_sql) == 3
    assert len(downgrade_sql) == 0

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == {"TestProductOne": {"missing_product_blocks_in_model": {"ProductBlockOneForTest"}}}

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()


def test_migrate_domain_models_remove_resource_type(
    test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, ProductTypeOneForTestProvisioning, _ = test_product_type_one
    _, ProductBlockOneForTestProvisioning, _ = test_product_block_one
    _, _, SubBlockOneForTest = test_product_sub_block_one

    class ProductBlockOneForTest(ProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        sub_block: SubBlockOneForTest
        sub_block_2: SubBlockOneForTest
        sub_block_list: List[SubBlockOneForTest]
        int_field: int
        str_field: str

    class ProductTypeOneForTestNew(ProductTypeOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_fixed_input: bool
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTestNew

    upgrade_sql, downgrade_sql = migrate_domain_models("example", True)

    assert len(upgrade_sql) == 4
    assert len(downgrade_sql) == 0

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == {
        "TestProductOne": {
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {"missing_resource_types_in_model": {"list_field"}}
            }
        }
    }

    for stmt in upgrade_sql:
        db.session.execute(stmt)
    db.session.commit()

    before_diff = ProductTypeOneForTestNew.diff_product_in_database(test_product_one)
    assert before_diff == {}

    for stmt in downgrade_sql:
        db.session.execute(stmt)
    db.session.commit()



from typing import Optional

import strawberry

from orchestrator.schemas.resource_type import ResourceTypeSchema


@strawberry.experimental.pydantic.type(model=ResourceTypeSchema)
class ResourceType:
    resource_type: strawberry.auto
    description: Optional[str]
    resource_type_id: strawberry.auto



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from dataclasses import dataclass
from datetime import datetime
from http import HTTPStatus
from shlex import shlex
from typing import Any, Generator, List, Optional, Union
from uuid import UUID

from more_itertools import chunked, first
from sqlalchemy import String, cast, func
from sqlalchemy.orm import Query
from sqlalchemy.sql import expression
from starlette.responses import Response
from structlog import get_logger

from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProcessTable, ProductTable, SubscriptionTable
from orchestrator.db.models import SubscriptionSearchView
from orchestrator.domain.base import SubscriptionModel

logger = get_logger(__name__)


def _quote_if_kv_pair(token: str) -> str:
    return f'"{token}"' if ":" in token else token


def _process_text_query(q: str) -> str:
    quote = '"'
    if q.count(quote) % 2 == 1:
        q += quote  # Add missing closing quote
    lex = shlex(q)
    lex.whitespace_split = True
    lex.quotes = '"'
    try:
        return " ".join([_quote_if_kv_pair(token) for token in lex])
    except ValueError:
        logger.debug("Error parsing text query.")
        return q


def _query_with_filters(
    response: Response,
    query: Query,
    range: Optional[List[int]] = None,
    sort: Optional[List[str]] = None,
    filters: Optional[List[str]] = None,
) -> List:
    if filters is not None:
        for filter in chunked(filters, 2):
            if filter and len(filter) == 2:
                field = filter[0]
                value = filter[1]
                value_as_bool = value.lower() in ("yes", "y", "ye", "true", "1", "ja", "insync")
                if value is not None:
                    if field.endswith("_gt"):
                        query = query.filter(SubscriptionTable.__dict__[field[:-3]] > value)
                    elif field.endswith("_gte"):
                        query = query.filter(SubscriptionTable.__dict__[field[:-4]] >= value)
                    elif field.endswith("_lte"):
                        query = query.filter(SubscriptionTable.__dict__[field[:-4]] <= value)
                    elif field.endswith("_lt"):
                        query = query.filter(SubscriptionTable.__dict__[field[:-3]] < value)
                    elif field.endswith("_ne"):
                        query = query.filter(SubscriptionTable.__dict__[field[:-3]] != value)
                    elif field == "insync":
                        query = query.filter(SubscriptionTable.insync.is_(value_as_bool))
                    elif field == "tags":
                        # For node and port selector form widgets
                        sub_values = value.split("-")
                        query = query.filter(func.lower(ProductTable.tag).in_([s.lower() for s in sub_values]))
                    elif field == "tag":
                        # For React table 7
                        sub_values = value.split("-")
                        query = query.filter(func.lower(ProductTable.tag).in_([s.lower() for s in sub_values]))
                    elif field == "product":
                        sub_values = value.split("-")
                        query = query.filter(func.lower(ProductTable.name).in_([s.lower() for s in sub_values]))
                    elif field == "status":
                        # For React table 7
                        statuses = value.split("-")
                        query = query.filter(SubscriptionTable.status.in_([s.lower() for s in statuses]))
                    elif field == "statuses":
                        # For port subscriptions
                        sub_values = value.split("-")
                        query = query.filter(SubscriptionTable.status.in_([s.lower() for s in sub_values]))
                    elif field == "organisation":
                        try:
                            value_as_uuid = UUID(value)
                        except (ValueError, AttributeError):
                            msg = "Not a valid customer_id, must be a UUID: '{value}'"
                            logger.debug(msg)
                            raise_status(HTTPStatus.BAD_REQUEST, msg)
                        query = query.filter(SubscriptionTable.customer_id == value_as_uuid)
                    elif field == "tsv":
                        # Quote key:value tokens. This will use the FOLLOWED BY operator (https://www.postgresql.org/docs/13/textsearch-controls.html)
                        processed_text_query = _process_text_query(value)

                        logger.debug("Running full-text search query:", value=processed_text_query)
                        # TODO: Make 'websearch_to_tsquery' into a sqlalchemy extension
                        query = query.join(SubscriptionSearchView).filter(
                            func.websearch_to_tsquery("simple", processed_text_query).op("@@")(
                                SubscriptionSearchView.tsv
                            )
                        )

                    elif field in SubscriptionTable.__dict__:
                        query = query.filter(cast(SubscriptionTable.__dict__[field], String).ilike("%" + value + "%"))

    if sort is not None and len(sort) >= 2:
        for item in chunked(sort, 2):
            if item and len(item) == 2:
                if item[0] in ["product", "tag"]:
                    field = "name" if item[0] == "product" else "tag"
                    if item[1].upper() == "DESC":
                        query = query.order_by(expression.desc(ProductTable.__dict__[field]))
                    else:
                        query = query.order_by(expression.asc(ProductTable.__dict__[field]))
                else:
                    if item[1].upper() == "DESC":
                        query = query.order_by(expression.desc(SubscriptionTable.__dict__[item[0]]))
                    else:
                        query = query.order_by(expression.asc(SubscriptionTable.__dict__[item[0]]))

    if range is not None and len(range) == 2:
        try:
            range_start = int(range[0])
            range_end = int(range[1])
            if range_start >= range_end:
                raise ValueError("range start must be lower than end")
        except (ValueError, AssertionError):
            msg = "Invalid range parameters"
            logger.exception(msg)
            raise_status(HTTPStatus.BAD_REQUEST, msg)
        total = query.count()
        query = query.slice(range_start, range_end)

        response.headers["Content-Range"] = f"subscriptions {range_start}-{range_end}/{total}"

    return query.all()


VALID_SORT_KEYS = {
    "creator": "created_by",
    "started": "started_at",
    "status": "last_status",
    "assignee": "assignee",
    "modified": "last_modified_at",
    "workflow": "workflow",
}


@dataclass
class ProductEnriched:
    product_id: UUID
    description: str
    name: str
    tag: str
    status: str
    product_type: str


@dataclass
class _Subscription:
    customer_id: UUID
    description: str
    end_date: float
    insync: bool
    start_date: float
    status: str
    subscription_id: UUID
    product: ProductEnriched


@dataclass
class _ProcessListItem:
    assignee: str
    created_by: Optional[str]
    failed_reason: Optional[str]
    last_modified_at: datetime
    pid: UUID
    started_at: datetime
    last_status: str
    last_step: Optional[str]
    subscriptions: List[_Subscription]
    workflow: str
    workflow_target: Optional[str]
    is_task: bool


def enrich_process(p: ProcessTable) -> _ProcessListItem:
    # p.subscriptions is a non JSON serializable AssociationProxy
    # So we need to build a list of Subscriptions here.
    subscriptions: List[_Subscription] = []
    for sub in p.subscriptions:
        prod = sub.product

        subscription = _Subscription(
            customer_id=sub.customer_id,
            description=sub.description,
            end_date=sub.end_date if sub.end_date else None,
            insync=sub.insync,
            start_date=sub.start_date if sub.start_date else None,
            status=sub.status,
            subscription_id=sub.subscription_id,
            product=ProductEnriched(
                product_id=prod.product_id,
                description=prod.description,
                name=prod.name,
                tag=prod.tag,
                status=prod.status,
                product_type=prod.product_type,
            ),
        )
        subscriptions.append(subscription)

    return _ProcessListItem(
        assignee=p.assignee,
        created_by=p.created_by,
        failed_reason=p.failed_reason,
        last_modified_at=p.last_modified_at,
        pid=p.pid,
        started_at=p.started_at.timestamp(),
        last_status=p.last_status,
        last_step=p.last_step,
        subscriptions=subscriptions,
        workflow=p.workflow,
        workflow_target=first([ps.workflow_target for ps in p.process_subscriptions], None),
        is_task=p.is_task,
    )


def update_in(dct: Union[dict, list], path: str, value: Any, sep: str = ".") -> None:
    """Update a value in a dict or list based on a path."""
    for x in path.split(sep):
        prev: Union[dict, list]
        if x.isdigit() and isinstance(dct, list):
            prev = dct
            dct = dct[int(x)]
        else:
            prev = dct
            dct = dict(dct).setdefault(x, {})
    prev[x] = value  # type: ignore


def get_in(dct: Union[dict, list], path: str, sep: str = ".") -> Any:
    """Get a value in a dict or list using the path and get the resulting key's value."""
    prev: Union[dict, list]
    for x in path.split(sep):
        if x.isdigit() and isinstance(dct, list):
            prev, dct = dct, dct[int(x)]
        else:
            prev, dct = dct, dict(dct).get(x)  # type: ignore
    return prev[x]  # type: ignore


def getattr_in(obj: Any, attr: str) -> Any:
    """Get an instance attribute value by path."""

    def _getattr(obj: object, attr: str) -> Any:
        if isinstance(obj, list):
            return obj[int(attr)]

        if isinstance(obj, dict):
            return obj.get(attr)

        return getattr(obj, attr, None)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def product_block_paths(subscription: Union[SubscriptionModel, dict]) -> List[str]:
    _subscription = subscription.dict() if isinstance(subscription, SubscriptionModel) else subscription

    def get_dict_items(d: dict) -> Generator:
        for k, v in d.items():
            if isinstance(v, dict):
                for k1, v1 in get_dict_items(v):
                    yield (f"{k}.{k1}", v1)
                yield (k, v)
            if isinstance(v, list):
                for index, list_item in enumerate(v):
                    if isinstance(list_item, dict):
                        for list_item_key, list_item_value in get_dict_items(list_item):
                            yield (f"{k}.{index}.{list_item_key}", list_item_value)
                        yield (f"{k}.{index}", list_item)

    return [path for path, value in get_dict_items(_subscription)]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import re
from typing import Dict, List, Optional, Tuple


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
class Cryptic:
    MAGIC = "$9$"
    MAGIC_SEARCH = r"\$9\$"
    FAMILY = ["QzF3n6/9CAtpu0O", "B1IREhcSyrleKvMW8LXx", "7N-dVbwsY2g4oaJZGUDj", "iHkq.mPf5T"]
    EXTRA: Dict[str, int] = {}
    VALID = ""
    NUM_ALPHA = ""
    ALPHA_NUM: Dict[str, int] = {}
    ENCODING = [[1, 4, 32], [1, 16, 32], [1, 8, 32], [1, 64], [1, 32], [1, 4, 16, 128], [1, 32, 64]]

    # ------------------------------------------------------------------------------
    def __init__(self) -> None:
        for fam in range(len(self.FAMILY)):
            for c in range(len(self.FAMILY[fam])):
                token = self.FAMILY[fam]
                self.EXTRA[token[c]] = 3 - fam

        self.NUM_ALPHA = "".join(self.FAMILY)
        self.VALID = "[" + self.MAGIC + self.NUM_ALPHA + "]"

        for num_alpha in range(len(self.NUM_ALPHA)):
            self.ALPHA_NUM[self.NUM_ALPHA[num_alpha]] = num_alpha

    # ------------------------------------------------------------------------------
    def _randc(self, count: int = 1) -> str:
        ret = ""

        while count > 0:
            ret += self.NUM_ALPHA[random.randint(0, len(self.NUM_ALPHA) - 1)]
            count -= 1

        return ret

    # ------------------------------------------------------------------------------
    def _gap_encode(self, pc: str, prev: str, enc: List[int]) -> str:
        literal_pc = ord(pc)
        crypt = ""
        gaps: List[int] = []

        for mod in reversed(enc):
            gaps.insert(0, int(literal_pc / mod))
            literal_pc %= mod

        for gap in gaps:
            gap += self.ALPHA_NUM[prev] + 1
            prev = self.NUM_ALPHA[gap % len(self.NUM_ALPHA)]
            c = prev
            crypt += c

        return crypt

    # ------------------------------------------------------------------------------
    def encrypt(self, plain: str, salt: Optional[str] = None) -> str:
        if salt is None:
            salt = self._randc(1)
        rand = self._randc(self.EXTRA[salt])
        pos = 0
        prev = salt
        crypt = self.MAGIC + str(salt) + str(rand)

        for p_index in range(len(plain)):
            p = plain[p_index]
            encode = self.ENCODING[pos % len(self.ENCODING)]
            crypt += self._gap_encode(p, prev, encode)
            prev = crypt[-1]
            pos += 1

        return crypt

    # ------------------------------------------------------------------------------
    def _nibble(self, cref: str, length: int) -> Tuple[str, str]:
        nib = cref[0:length]
        cref = cref[length:]

        return nib, cref

    # ------------------------------------------------------------------------------
    def _gap(self, c1: str, c2: str) -> int:
        return ((self.ALPHA_NUM[c2] - self.ALPHA_NUM[c1]) % len(self.NUM_ALPHA)) - 1

    # ------------------------------------------------------------------------------
    def _gap_decode(self, gaps: List[int], dec: List[int]) -> str:
        num = 0
        assert len(gaps) == len(dec)
        for i in range(len(gaps)):
            num += gaps[i] * dec[i]

        return chr(num % 256)

    # ------------------------------------------------------------------------------
    def decrypt(self, crypt: Optional[str]) -> Optional[str]:
        if crypt is None or len(crypt) == 0:
            print("Invalid Crypt")
            return None

        valid_chars = re.compile(self.VALID)
        if valid_chars.match(crypt) is not None:
            match_object = re.match(self.MAGIC_SEARCH, crypt)
            assert match_object
            chars = crypt[match_object.end() :]
            first, chars = self._nibble(chars, 1)
            var, chars = self._nibble(chars, self.EXTRA[first])
            prev = first
            decrypt_str = ""

            while len(chars) > 0:
                decode = self.ENCODING[len(decrypt_str) % len(self.ENCODING)]
                length = len(decode)
                nibble, chars = self._nibble(chars, length)
                gaps = []
                for i in range(len(nibble)):
                    gaps.append(self._gap(prev, nibble[i]))
                    prev = nibble[i]

                decrypt_str += self._gap_decode(gaps, decode)

            return decrypt_str
        else:
            print(crypt + " is invalid !!")
            return None


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    cryptic = Cryptic()
    print(cryptic.decrypt(cryptic.encrypt("abcd1234")))
    print(cryptic.decrypt(cryptic.encrypt("no")))
    print(cryptic.decrypt(cryptic.encrypt("Ramesh")))
    print(cryptic.encrypt("bgp password for SAP1"))
    print(cryptic.encrypt("bgp password for SAP2"))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer

from orchestrator.cli import database, generate, scheduler

app = typer.Typer()
app.add_typer(scheduler.app, name="scheduler", help="Access all the scheduler functions")
app.add_typer(database.app, name="db", help="Interact with the application database")
app.add_typer(generate.app, name="generate", help="Generate products, workflows and other artifacts")


if __name__ == "__main__":
    app()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import timedelta

from orchestrator.db import ProcessTable, db
from orchestrator.targets import Target
from orchestrator.types import State
from orchestrator.utils.datetime import nowtz
from orchestrator.workflow import ProcessStatus, StepList, done, init, step, workflow


@step("Clean up tasks older than one week")
def remove_tasks() -> State:
    delta = timedelta(days=3)
    now = nowtz()
    three_days_ago = now - delta
    tasks = (
        ProcessTable.query.filter(ProcessTable.is_task.is_(True))
        .filter(ProcessTable.last_status == ProcessStatus.COMPLETED)
        .filter(ProcessTable.last_modified_at <= three_days_ago)
        .all()
    )
    count = 0
    for task in tasks:
        db.session.delete(task)
        count += 1

    return {"tasks_removed": count}


@workflow("Clean up old tasks", target=Target.SYSTEM)
def task_clean_up_tasks() -> StepList:
    return init >> remove_tasks >> done



from typing import Union

import structlog

from orchestrator.db.filters import Filter
from orchestrator.db.filters.product import filter_products
from orchestrator.db.models import ProductTable
from orchestrator.db.range.range import apply_range_to_query
from orchestrator.db.sorting.product import sort_products
from orchestrator.db.sorting.sorting import Sort
from orchestrator.graphql.pagination import Connection, PageInfo
from orchestrator.graphql.schemas.product import ProductType
from orchestrator.graphql.types import CustomInfo, GraphqlFilter, GraphqlSort
from orchestrator.graphql.utils.create_resolver_error_handler import create_resolver_error_handler

logger = structlog.get_logger(__name__)


async def resolve_products(
    info: CustomInfo,
    filter_by: Union[list[GraphqlFilter], None] = None,
    sort_by: Union[list[GraphqlSort], None] = None,
    first: int = 10,
    after: int = 0,
) -> Connection[ProductType]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.info("resolve_products() called", range=[after, after + first], sort=sort_by, filter=pydantic_filter_by)

    query = filter_products(ProductTable.query, pydantic_filter_by, _error_handler)
    query = sort_products(query, pydantic_sort_by, _error_handler)
    total = query.count()
    query = apply_range_to_query(query, after, first)

    products = query.all()
    has_next_page = len(products) > first

    products = products[:first]
    products_length = len(products)
    start_cursor = after if products_length else None
    end_cursor = after + products_length - 1
    page_products = [ProductType.from_pydantic(p) for p in products]

    return Connection(
        page=page_products,
        page_info=PageInfo(
            has_previous_page=bool(after),
            has_next_page=has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            total_items=total if total else None,
        ),
    )



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from http import HTTPStatus
from typing import Any, Callable, Dict, Generator, List, Literal, Optional, Tuple, Type, TypedDict, TypeVar, Union
from uuid import UUID

import strawberry
from pydantic import BaseModel
from pydantic.typing import get_args, get_origin, is_union

UUIDstr = str
State = Dict[str, Any]
JSON = Any
# ErrorState is either a string containing an error message, a catched Exception or a tuple containing a message and
# a HTTP status code
ErrorState = Union[str, Exception, Tuple[str, Union[int, HTTPStatus]]]
# An ErrorDict should have the following keys:
# error: str  # A message describing the error
# class: str[Optional]  # The exception class name (type)
# status_code: Optional[int]  # HTTP Status code (optional)
# traceback: Optional[str]  # A python traceback as a string formatted by nwastdlib.ex.show_ex
ErrorDict = Dict[str, Union[str, int, List[Dict[str, Any]], "InputForm", None]]
StateStepFunc = Callable[[State], State]
StepFunc = Callable[..., Optional[State]]
BroadcastFunc = Callable[[UUID, Dict], None]


class strEnum(str, Enum):
    def __str__(self) -> str:
        return self.value

    @classmethod
    def values(cls) -> List:
        return [obj.value for obj in cls]


@strawberry.enum
class SubscriptionLifecycle(strEnum):
    INITIAL = "initial"
    ACTIVE = "active"
    MIGRATING = "migrating"
    DISABLED = "disabled"
    TERMINATED = "terminated"
    PROVISIONING = "provisioning"


@strawberry.enum
class AcceptItemType(strEnum):
    INFO = "info"
    LABEL = "label"
    WARNING = "warning"
    URL = "url"
    CHECKBOX = "checkbox"
    SUBCHECKBOX = ">checkbox"
    OPTIONAL_CHECKBOX = "checkbox?"
    OPTIONAL_SUBCHECKBOX = ">checkbox?"
    SKIP = "skip"
    VALUE = "value"
    MARGIN = "margin"


# TODO #1321: old code that protected against unsafe changes in subs
# The key is the Parent subscription life cycle status. The keys are lists of safe transitions for child subscriptions.
SAFE_USED_BY_TRANSITIONS_FOR_STATUS = {
    SubscriptionLifecycle.INITIAL: [
        SubscriptionLifecycle.INITIAL,
    ],
    SubscriptionLifecycle.ACTIVE: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.PROVISIONING,
        SubscriptionLifecycle.MIGRATING,
        SubscriptionLifecycle.ACTIVE,
        SubscriptionLifecycle.TERMINATED,
        SubscriptionLifecycle.DISABLED,
    ],
    SubscriptionLifecycle.MIGRATING: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.MIGRATING,
        SubscriptionLifecycle.TERMINATED,
    ],
    SubscriptionLifecycle.PROVISIONING: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.PROVISIONING,
        SubscriptionLifecycle.ACTIVE,
        SubscriptionLifecycle.TERMINATED,
    ],
    SubscriptionLifecycle.TERMINATED: [SubscriptionLifecycle.INITIAL, SubscriptionLifecycle.TERMINATED],
    SubscriptionLifecycle.DISABLED: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.DISABLED,
        SubscriptionLifecycle.TERMINATED,
    ],
}


AcceptData = List[Union[Tuple[str, AcceptItemType], Tuple[str, AcceptItemType, Dict]]]

InputForm = Type[BaseModel]

T = TypeVar("T", bound=BaseModel)
FormGenerator = Generator[Type[T], T, State]
SimpleInputFormGenerator = Callable[..., InputForm]
InputFormGenerator = Callable[..., FormGenerator]
InputStepFunc = Union[SimpleInputFormGenerator, InputFormGenerator]
StateSimpleInputFormGenerator = Callable[[State], InputForm]
StateInputFormGenerator = Callable[[State], FormGenerator]
StateInputStepFunc = Union[StateSimpleInputFormGenerator, StateInputFormGenerator]
SubscriptionMapping = Dict[str, List[Dict[str, str]]]


class SummaryData(TypedDict, total=False):
    headers: List[str]
    labels: List[str]
    columns: List[List[Union[str, int, bool, float]]]


def is_of_type(t: Any, test_type: Any) -> bool:
    """Check if annotation type is valid for type.

    >>> is_of_type(list, list)
    True
    >>> is_of_type(List[int], List[int])
    True
    >>> is_of_type(strEnum, str)
    True
    >>> is_of_type(strEnum, Enum)
    True
    >>> is_of_type(int, str)
    False
    """

    if is_union_type(test_type):
        return any(get_origin(t) is get_origin(arg) for arg in get_args(test_type))

    if (
        get_origin(t)
        and get_origin(test_type)
        and get_origin(t) is get_origin(test_type)
        and get_args(t) == get_args(test_type)
    ):
        return True

    if test_type is t:
        # Test type is a typing type instance and matches
        return True

    # Workaround for the fact that you can't call issubclass on typing types
    try:
        return issubclass(t, test_type)
    except TypeError:
        return False


# TODO #1330: Fix comparison of is_of_type(Union[str, int], Union[int, str]), it returns False but it should be True
def is_list_type(t: Any, test_type: Optional[type] = None) -> bool:
    """Check if `t` is list type.

    And optionally check if the list items are of `test_type`

    >>> is_list_type(List[int])
    True
    >>> is_list_type(Optional[List[int]])
    True
    >>> is_list_type(Optional[List[int]], int)
    True
    >>> is_list_type(Optional[List[int]], str)
    False
    >>> is_list_type(Optional[int])
    False
    >>> is_list_type(List[Tuple[int, int]])
    True
    >>> is_list_type(List[Tuple[int, int]], int)
    False
    >>> is_list_type(List[Tuple[int, int]], Tuple[int, int])
    True
    >>> is_list_type(List[strEnum], Enum)
    True
    >>> is_list_type(int)
    False
    >>> is_list_type(Literal[1,2,3])
    False
    >>> is_list_type(List[Union[str, int]])
    True
    >>> is_list_type(List[Union[str, int]], Union[str, int])
    False
    >>> is_list_type(List[Union[str, int]], str)
    True
    >>> is_list_type(List[Union[str, int]], int)
    False
    >>> is_list_type(List[Union[str, int]], Union[int, int])
    False
    """
    if get_origin(t):
        if is_optional_type(t) or is_union_type(t):
            for arg in get_args(t):
                if is_list_type(arg, test_type):
                    return True
        elif get_origin(t) == Literal:  # type:ignore
            return False  # Literal cannot contain lists see pep 586
        elif issubclass(get_origin(t), list):  # type: ignore
            if test_type and get_args(t):
                first_arg = get_args(t)[0]
                # To support a list with union of multiple product blocks.
                if is_union_type(first_arg) and get_args(first_arg) and not is_union_type(test_type):
                    first_arg = get_args(first_arg)[0]
                return is_of_type(first_arg, test_type)
            else:
                return True

    return False


def is_optional_type(t: Any, test_type: Optional[type] = None) -> bool:
    """Check if `t` is optional type (Union[None, ...]).

    And optionally check if T is of `test_type`

    >>> is_optional_type(Optional[int])
    True
    >>> is_optional_type(Union[None, int])
    True
    >>> is_optional_type(Union[int, str, None])
    True
    >>> is_optional_type(Optional[int], int)
    True
    >>> is_optional_type(Optional[int], str)
    False
    >>> is_optional_type(Optional[State], int)
    False
    >>> is_optional_type(Optional[State], State)
    True
    """
    if is_union(get_origin(t)) and None.__class__ in get_args(t):
        for arg in get_args(t):
            if arg is None.__class__:
                continue

            return not test_type or is_of_type(arg, test_type)
    return False


def is_union_type(t: Any, test_type: Optional[type] = None) -> bool:
    """Check if `t` is union type (Union[Type, AnotherType]).

    Optionally check if T is of `test_type` We cannot check for literal Nones.

    >>> is_union_type(Union[int, str])
    True
    >>> is_union_type(Union[int, str], str)
    True
    >>> is_union_type(Union[int, str], Union[int, str])
    True
    >>> is_union_type(Union[int, None])
    True
    >>> is_union_type(int)
    False
    """
    if not is_union(get_origin(t)):
        return False
    if not test_type:
        return True

    if is_of_type(t, test_type):
        return True
    for arg in get_args(t):
        result = is_of_type(arg, test_type)
        if result:
            return result
    return False


def get_possible_product_block_types(list_field_type: Any) -> dict:
    possible_product_block_types = {}
    if is_union_type(list_field_type):
        for list_item_field_type in get_args(list_field_type):
            if (
                not isinstance(None, list_item_field_type)
                and list_item_field_type.name not in possible_product_block_types
            ):
                possible_product_block_types[list_item_field_type.name] = list_item_field_type
    else:
        possible_product_block_types = {list_field_type.name: list_field_type}
    return possible_product_block_types



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import structlog
from sqlalchemy import func

from orchestrator.api.helpers import _process_text_query
from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters.filters import generic_filter
from orchestrator.db.filters.generic_filters import (
    generic_bool_filter,
    generic_is_like_filter,
    generic_range_filters,
    generic_values_in_column_filter,
)
from orchestrator.db.models import SubscriptionSearchView

logger = structlog.get_logger(__name__)


def tsv_filter(query: SearchQuery, value: str) -> SearchQuery:
    # Quote key:value tokens. This will use the FOLLOWED BY operator (https://www.postgresql.org/docs/13/textsearch-controls.html)
    processed_text_query = _process_text_query(value)

    logger.debug("Running full-text search query:", value=processed_text_query)
    # TODO: Make 'websearch_to_tsquery' into a sqlalchemy extension
    return query.join(SubscriptionSearchView).filter(
        func.websearch_to_tsquery("simple", processed_text_query).op("@@")(SubscriptionSearchView.tsv)
    )


def subscription_list_filter(query: SearchQuery, value: str) -> SearchQuery:
    values = [s.lower() for s in value.split(",")]
    return query.filter(SubscriptionTable.subscription_id.in_(values))


status_filter = generic_values_in_column_filter(SubscriptionTable.status)
product_filter = generic_values_in_column_filter(ProductTable.name)
tags_filter = generic_values_in_column_filter(ProductTable.tag)

start_date_range_filters = generic_range_filters(SubscriptionTable.start_date)
end_date_range_filters = generic_range_filters(SubscriptionTable.end_date)


SUBSCRIPTION_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = (
    {
        "subscriptionId": generic_is_like_filter(SubscriptionTable.subscription_id),
        "subscriptionIds": subscription_list_filter,
        "description": generic_is_like_filter(SubscriptionTable.description),
        "status": status_filter,
        "product": product_filter,
        "insync": generic_bool_filter(SubscriptionTable.insync),
        "note": generic_is_like_filter(SubscriptionTable.note),
        "statuses": status_filter,
        "tags": tags_filter,
        "tag": tags_filter,
        "tsv": tsv_filter,
        "startDate": generic_is_like_filter(SubscriptionTable.start_date),
        "endDate": generic_is_like_filter(SubscriptionTable.end_date),
    }
    | start_date_range_filters
    | end_date_range_filters
)


filter_subscriptions = generic_filter(SUBSCRIPTION_FILTER_FUNCTIONS_BY_COLUMN)



#!/usr/bin/env python3

"""Create migration for models based on diff_product_in_database.

$ PYTHONPATH=. python bin/list_workflows

"""
import sys
from typing import Dict, List, Optional, Set, Tuple, Type, Union
from uuid import UUID

import structlog

from orchestrator.cli.domain_gen_helpers.fixed_input_helpers import (
    generate_create_fixed_inputs_sql,
    generate_delete_fixed_inputs_sql,
    generate_update_fixed_inputs_sql,
    map_update_fixed_inputs,
)
from orchestrator.cli.domain_gen_helpers.helpers import (
    map_create_fixed_inputs,
    map_create_product_block_relations,
    map_create_resource_type_relations,
    map_delete_fixed_inputs,
    map_delete_product_block_relations,
    map_delete_resource_type_relations,
)
from orchestrator.cli.domain_gen_helpers.product_block_helpers import (
    generate_create_product_block_instance_relations_sql,
    generate_create_product_block_relations_sql,
    generate_create_product_blocks_sql,
    generate_delete_product_block_relations_sql,
    generate_delete_product_blocks_sql,
    map_create_product_blocks,
    map_delete_product_blocks,
    map_product_block_additional_relations,
)
from orchestrator.cli.domain_gen_helpers.product_helpers import (
    generate_create_product_instance_relations_sql,
    generate_create_product_relations_sql,
    generate_create_products_sql,
    generate_delete_product_relations_sql,
    generate_delete_products_sql,
    map_product_additional_relations,
)
from orchestrator.cli.domain_gen_helpers.resource_type_helpers import (
    generate_create_resource_type_instance_values_sql,
    generate_create_resource_type_relations_sql,
    generate_create_resource_types_sql,
    generate_delete_resource_type_relations_sql,
    generate_delete_resource_types_sql,
    generate_rename_resource_types_sql,
    generate_update_resource_type_block_relations_sql,
    generate_update_resource_type_instance_values_sql,
    map_create_resource_type_instances,
    map_create_resource_types,
    map_delete_resource_types,
    map_rename_resource_types,
    map_update_product_block_resource_types,
)
from orchestrator.cli.domain_gen_helpers.types import DomainModelChanges, ModelUpdates
from orchestrator.cli.helpers.input_helpers import get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, print_fmt
from orchestrator.db.models import ProductTable
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductBlockModel, SubscriptionModel, get_depends_on_product_block_type_list
from orchestrator.settings import app_settings

logger = structlog.get_logger(__name__)


def should_skip(product_name: str) -> bool:
    return any(name in product_name.lower() for name in app_settings.SKIP_MODEL_FOR_MIGRATION_DB_DIFF)


def map_product_blocks_in_class(
    model_class: Union[Type[SubscriptionModel], Type[ProductBlockModel]],
    product_blocks: Dict[str, Type[ProductBlockModel]],
) -> Dict[str, Type[ProductBlockModel]]:
    """Create mapping of all existing product block models related to a model.

    Args:
        - model_class: a product class (SubscriptionModel) or product block class (ProductBlockModel).

    Returns a Dict with product block models by product block name.
    """
    product_blocks_types_in_model = get_depends_on_product_block_type_list(
        model_class._get_depends_on_product_block_types()
    )

    new_blocks = {block.name: block for block in product_blocks_types_in_model if block.name not in product_blocks}
    product_blocks = {**product_blocks, **new_blocks}

    blocks_map = {
        name: block_cls
        for block in new_blocks.values()
        for name, block_cls in map_product_blocks_in_class(block, product_blocks).items()
        if block_cls.name not in product_blocks
    }
    return {**product_blocks, **blocks_map}


def map_product_blocks(product_classes: List[Type[SubscriptionModel]]) -> Dict[str, Type[ProductBlockModel]]:
    """Create mapping of all existing product block models related to products.

    Args:
        - product_classes: List of product classes.

    Returns a Dict with product block models by product block name.
    """

    product_blocks: Dict[str, Type[ProductBlockModel]] = {}
    for product_class in product_classes:
        product_blocks = {**product_blocks, **map_product_blocks_in_class(product_class, product_blocks)}
    return product_blocks


def map_differences_unique(
    registered_products: Dict[str, Type[SubscriptionModel]], existing_products: List[Tuple[str, UUID]]
) -> Dict[str, Dict[str, Dict[str, Set[str]]]]:
    """Create a unique map for products and product block differences from the database.

    Args:
        - registered_products: Dict with product models by product name.
        - existing_products: List with tuples of product name and its uuid that exist in the database.

    Returns a dict with product and product block differences from the database without duplicates.
    """
    model_diffs: Dict[str, Dict[str, Dict[str, Set[str]]]] = {"products": {}, "blocks": {}}

    for product_name, product_id in existing_products:
        if should_skip(product_name) or product_name not in registered_products:
            continue

        product_class = registered_products[product_name]
        if diff := product_class.diff_product_in_database(product_id):
            model_diffs["products"][product_name] = {k: v for k, v in diff[product_name].items() if isinstance(v, set)}

            if missing_in_depends_on_blocks := diff[product_name].get("missing_in_depends_on_blocks"):
                # since missing_in_depends_on_blocks is always a dict the type is ignored.
                new_model_diffs: Dict[str, Dict[str, Set[str]]] = {
                    name: diff  # type: ignore
                    for name, diff in missing_in_depends_on_blocks.items()  # type: ignore
                    if name not in model_diffs["blocks"]
                }
                model_diffs["blocks"] = {**model_diffs["blocks"], **new_model_diffs}
    return model_diffs


def remove_updated_properties(
    updates: ModelUpdates, model_diffs: Dict[str, Dict[str, Dict[str, Set[str]]]]
) -> Dict[str, Dict[str, Dict[str, Set[str]]]]:
    for product_name, updated_fixed_inputs in updates.fixed_inputs.items():
        product_diffs = model_diffs["products"][product_name]
        product_diffs["missing_fixed_inputs_in_model"] = product_diffs.get(
            "missing_fixed_inputs_in_model", set()
        ) - set(updated_fixed_inputs.keys())
        product_diffs["missing_fixed_inputs_in_db"] = product_diffs.get("missing_fixed_inputs_in_db", set()) - set(
            updated_fixed_inputs.values()
        )
        model_diffs["products"][product_name] = product_diffs
    for block_name, diffs in model_diffs["blocks"].items():
        missing_rt_in_model = diffs.get("missing_resource_types_in_model", set())
        missing_rt_in_db = diffs.get("missing_resource_types_in_db", set())

        rt_updates = {
            k: v for k, v in updates.resource_types.items() if k in missing_rt_in_model and v in missing_rt_in_db
        }
        updated_old_rts = set(rt_updates.keys())
        updated_new_rts = set(rt_updates.values())
        if block_updates := updates.block_resource_types.get(block_name, {}):
            updated_old_rts = updated_old_rts | set(block_updates.keys())
            updated_new_rts = updated_new_rts | set(block_updates.values())
        diffs["missing_resource_types_in_model"] = missing_rt_in_model - updated_old_rts
        diffs["missing_resource_types_in_db"] = missing_rt_in_db - updated_new_rts
        model_diffs["blocks"][block_name] = diffs
    return model_diffs


def map_changes(
    model_diffs: Dict[str, Dict[str, Dict[str, Set[str]]]],
    products: Dict[str, Type[SubscriptionModel]],
    product_blocks: Dict[str, Type[ProductBlockModel]],
    db_product_names: List[str],
    inputs: Dict[str, Dict[str, str]],
    updates: Optional[ModelUpdates],
) -> DomainModelChanges:
    """Map changes that need to be made to fix differences between models and database.

    Args:
        - model_diffs: Dict with product and product block differences.
            - products: Dict with product differences.
            - blocks: Dict with product block differences.
        - products: Dict with product model by product name.
        - product_blocks: Dict with product block model by product name.
        - db_product_names: Product names out of the database.
        - inputs: Optional Dict with prefilled values.
        - updates: Optional Dict.

    Returns: Mapped changes.
    """
    create_products = {name: model for name, model in products.items() if name not in db_product_names}
    delete_products = {name for name in db_product_names if name not in SUBSCRIPTION_MODEL_REGISTRY}

    # updates need to go before create or deletes.
    if not updates:
        renamed_resource_types = map_rename_resource_types(model_diffs["blocks"], product_blocks)
        update_block_resource_types = map_update_product_block_resource_types(
            model_diffs["blocks"], renamed_resource_types
        )
        updates = ModelUpdates(
            fixed_inputs=map_update_fixed_inputs(model_diffs["products"]),
            resource_types=renamed_resource_types,
            block_resource_types=update_block_resource_types,
        )

    model_diffs = remove_updated_properties(updates, model_diffs)
    create_resource_type_relations = map_create_resource_type_relations(model_diffs["blocks"])
    delete_resource_type_relations = map_delete_resource_type_relations(model_diffs["blocks"])

    changes = DomainModelChanges(
        create_products=create_products,
        delete_products=delete_products,
        create_product_fixed_inputs=map_create_fixed_inputs(model_diffs["products"]),
        update_product_fixed_inputs=updates.fixed_inputs,
        delete_product_fixed_inputs=map_delete_fixed_inputs(model_diffs["products"]),
        create_product_to_block_relations=map_create_product_block_relations(model_diffs["products"]),
        delete_product_to_block_relations=map_delete_product_block_relations(model_diffs["products"]),
        rename_resource_types=updates.resource_types,
        update_block_resource_types=updates.block_resource_types,
        delete_resource_types=map_delete_resource_types(
            delete_resource_type_relations, list(updates.resource_types.keys()), product_blocks
        ),
        create_resource_type_relations=create_resource_type_relations,
        delete_resource_type_relations=delete_resource_type_relations,
        create_product_blocks=map_create_product_blocks(product_blocks),
        delete_product_blocks=map_delete_product_blocks(product_blocks),
        create_product_block_relations=map_create_product_block_relations(model_diffs["blocks"]),
        delete_product_block_relations=map_delete_product_block_relations(model_diffs["blocks"]),
    )

    changes = map_product_additional_relations(changes)
    changes = map_product_block_additional_relations(changes)
    temp = {key for v in changes.update_block_resource_types.values() for key in v.values()}
    related_resource_type_names = set(changes.create_resource_type_relations.keys()) | temp
    existing_renamed_rts = set(changes.rename_resource_types.values())
    changes.create_resource_types = map_create_resource_types(related_resource_type_names, existing_renamed_rts)
    changes.create_resource_type_instance_relations = map_create_resource_type_instances(changes)

    return changes


def generate_upgrade_sql(changes: DomainModelChanges, inputs: Dict[str, Dict[str, str]]) -> List[str]:
    """Generate upgrade SQL with mapped changes.

    Args:
        - changes: Mapping of model changes.
        - inputs: Optional Dict with prefilled values.

    Returns: List of SQL strings to upgrade the database.
    """
    return (
        generate_update_fixed_inputs_sql(changes.update_product_fixed_inputs)
        + generate_rename_resource_types_sql(changes.rename_resource_types)
        + generate_delete_resource_type_relations_sql(changes.delete_resource_type_relations)
        + generate_delete_product_block_relations_sql(changes.delete_product_block_relations)
        + generate_delete_product_relations_sql(changes.delete_product_to_block_relations)
        + generate_delete_resource_types_sql(changes.delete_resource_types)
        + generate_delete_product_blocks_sql(changes.delete_product_blocks)
        + generate_delete_fixed_inputs_sql(changes.delete_product_fixed_inputs)
        + generate_delete_products_sql(changes.delete_products)
        + generate_create_products_sql(changes.create_products, inputs)
        + generate_create_fixed_inputs_sql(changes.create_product_fixed_inputs, inputs)
        + generate_create_product_blocks_sql(changes.create_product_blocks, inputs)
        + generate_create_resource_types_sql(changes.create_resource_types, inputs)
        + generate_update_resource_type_block_relations_sql(changes.update_block_resource_types)
        + generate_update_resource_type_instance_values_sql(changes.update_block_resource_types)
        + generate_create_product_relations_sql(changes.create_product_to_block_relations)
        + generate_create_product_block_relations_sql(changes.create_product_block_relations)
        + generate_create_resource_type_relations_sql(changes.create_resource_type_relations)
        + generate_create_product_instance_relations_sql(changes.create_product_to_block_relations)
        + generate_create_product_block_instance_relations_sql(changes.create_product_block_relations)
        + generate_create_resource_type_instance_values_sql(changes.create_resource_type_instance_relations, inputs)
    )


def generate_downgrade_sql(changes: DomainModelChanges) -> List[str]:
    """Generate downgrade SQL with mapped changes.

    Does not revert deleted subscription instances and subscription instance values!

    Args:
        - changes: Mapping of model changes.

    Returns: List of SQL strings to downgrade the database back before upgrade SQL.
    """
    sql_revert_create_fixed_inputs = generate_delete_fixed_inputs_sql(changes.create_product_fixed_inputs)

    update_revert_map = {
        name: {new: old for old, new in updates.items()}
        for name, updates in changes.update_product_fixed_inputs.items()
    }
    sql_revert_update_fixed_inputs = generate_update_fixed_inputs_sql(update_revert_map)
    sql_revert_delete_fixed_inputs = generate_create_fixed_inputs_sql(
        changes.delete_product_fixed_inputs, {}, revert=True
    )

    sql_revert_create_resource_type_relations = generate_delete_resource_type_relations_sql(
        changes.create_resource_type_relations
    )
    sql_revert_create_resource_types = generate_delete_resource_types_sql(changes.create_resource_types)
    sql_revert_rename_resource_types = generate_rename_resource_types_sql(
        {new: old for old, new in changes.rename_resource_types.items()}
    )
    reversed_update_block_resource_types = {
        block: {new: old for old, new in rt_updates.items()}
        for block, rt_updates in changes.update_block_resource_types.items()
    }
    sql_revert_update_block_resource_types = generate_update_resource_type_block_relations_sql(
        reversed_update_block_resource_types
    )
    sql_revert_update_block_instance_values = generate_update_resource_type_instance_values_sql(
        reversed_update_block_resource_types
    )

    sql_revert_create_product_product_block_relations = generate_delete_product_relations_sql(
        changes.create_product_to_block_relations,
    )
    sql_revert_create_product_block_depends_blocks = generate_delete_product_block_relations_sql(
        changes.create_product_block_relations
    )

    sql_revert_create_product_blocks = generate_delete_product_blocks_sql(set(changes.create_product_blocks.keys()))
    sql_revert_create_products = generate_delete_products_sql(set(changes.create_products.keys()))

    return (
        sql_revert_create_resource_type_relations
        + sql_revert_update_block_resource_types
        + sql_revert_update_block_instance_values
        + sql_revert_create_resource_types
        + sql_revert_create_product_product_block_relations
        + sql_revert_create_product_block_depends_blocks
        + sql_revert_create_fixed_inputs
        + sql_revert_create_product_blocks
        + sql_revert_create_products
        + sql_revert_delete_fixed_inputs
        + sql_revert_rename_resource_types
        + sql_revert_update_fixed_inputs
    )


def create_domain_models_migration_sql(
    inputs: Dict[str, Dict[str, str]],
    updates: Optional[ModelUpdates],
    is_test: bool = False,
) -> Tuple[List[str], List[str]]:
    """Create tuple with list for upgrade and downgrade SQL statements based on SubscriptionModel.diff_product_in_database.

    You will be prompted with inputs for new models and resource type updates.

    Args:
        - inputs: dict with pre-defined input values

    Returns tuple:
        - list of upgrade SQL statements in string format.
        - list of downgrade SQL statements in string format.
    """
    existing_products: List[Tuple[str, UUID]] = list(
        ProductTable.query.with_entities(ProductTable.name, ProductTable.product_id)
    )
    db_product_names: List[str] = [product_name for product_name, _ in existing_products]

    products = SUBSCRIPTION_MODEL_REGISTRY
    product_blocks = map_product_blocks(list(SUBSCRIPTION_MODEL_REGISTRY.values()))
    model_diffs = map_differences_unique(products, existing_products)

    changes = map_changes(model_diffs, products, product_blocks, db_product_names, inputs, updates)

    logger.info("create_products", create_products=changes.create_products)
    logger.info("delete_products", delete_products=changes.delete_products)
    logger.info("create_product_fixed_inputs", create_product_fixed_inputs=changes.create_product_fixed_inputs)
    logger.info("update_product_fixed_inputs", update_product_fixed_inputs=changes.update_product_fixed_inputs)
    logger.info("delete_product_fixed_inputs", delete_product_fixed_inputs=changes.delete_product_fixed_inputs)
    logger.info(
        "create_product_to_block_relations", create_product_to_block_relations=changes.create_product_to_block_relations
    )
    logger.info(
        "delete_product_to_block_relations", delete_product_to_block_relations=changes.delete_product_to_block_relations
    )
    logger.info("create_resource_types", create_resource_types=changes.create_resource_types)
    logger.info("rename_resource_types", rename_resource_types=changes.rename_resource_types)
    logger.info("update_block_resource_types", update_block_resource_types=changes.update_block_resource_types)
    logger.info("delete_resource_types", delete_resource_types=changes.delete_resource_types)
    logger.info("create_resource_type_relations", create_resource_type_relations=changes.create_resource_type_relations)
    logger.info(
        "create_resource_type_instance_relations",
        create_resource_type_instance_relations=changes.create_resource_type_instance_relations,
    )
    logger.info("delete_resource_type_relations", delete_resource_type_relations=changes.delete_resource_type_relations)
    logger.info("create_product_blocks", create_blocks=changes.create_product_blocks)
    logger.info("delete_product_blocks", delete_blocks=changes.delete_product_blocks)
    logger.info("create_product_block_relations", create_product_block_relations=changes.create_product_block_relations)
    logger.info("delete_product_block_relations", delete_product_block_relations=changes.delete_product_block_relations)

    print_fmt("\nWARNING:", flags=[COLOR.BOLD, COLOR.YELLOW], end=" ")
    print_fmt("Deleting products will also delete its subscriptions.")

    if not is_test and "y" not in get_user_input("Confirm the above actions [y/n]: ", "n").lower():
        sys.exit()

    sql_upgrade_stmts = generate_upgrade_sql(changes, inputs)
    sql_downgrade_stmts = generate_downgrade_sql(changes)

    return sql_upgrade_stmts, sql_downgrade_stmts



import pytest

from orchestrator.services.subscriptions import get_subscription
from test.unit_tests.workflows import assert_complete, extract_state, run_workflow

TEST = "Some note"


@pytest.mark.workflow
def test_modify_note(responses, generic_subscription_1):
    init_state = [{"subscription_id": generic_subscription_1}, {"note": TEST}]

    result, process, step_log = run_workflow("modify_note", init_state)
    assert_complete(result)

    # assert state for correctness
    state = extract_state(result)
    assert state["old_note"] is None
    assert state["note"] == TEST
    assert state["__old_subscriptions__"].get(generic_subscription_1)
    assert state["__old_subscriptions__"][generic_subscription_1]["note"] is None
    assert state["__old_subscriptions__"][generic_subscription_1]["description"] == "Generic Subscription One"

    # assert subscription for correctness
    subscription = get_subscription(generic_subscription_1)
    assert subscription.note == TEST



from typing import Optional
from uuid import uuid4

import pytest

from orchestrator.db import ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_type_sub_list_union(test_product_block_with_list_union):
    (
        ProductBlockWithListUnionForTestInactive,
        ProductBlockWithListUnionForTestProvisioning,
        ProductBlockWithListUnionForTest,
    ) = test_product_block_with_list_union

    class ProductSubListUnionInactive(SubscriptionModel, is_base=True):
        test_block: Optional[ProductBlockWithListUnionForTestInactive]

    class ProductSubListUnionProvisioning(ProductSubListUnionInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        test_block: ProductBlockWithListUnionForTestProvisioning

    class ProductSubListUnion(ProductSubListUnionProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: ProductBlockWithListUnionForTest

    SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"] = ProductSubListUnion
    yield ProductSubListUnionInactive, ProductSubListUnionProvisioning, ProductSubListUnion
    del SUBSCRIPTION_MODEL_REGISTRY["ProductSubListUnion"]


@pytest.fixture
def test_product_sub_list_union(test_product_block_with_list_union_db):
    product = ProductTable(
        name="ProductSubListUnion",
        description="Product with Union sub product_block",
        tag="UnionSub",
        product_type="Test",
        status="active",
    )
    product_block, _, _ = test_product_block_with_list_union_db
    product.product_blocks = [product_block]
    db.session.add(product)
    db.session.commit()

    return product.product_id


@pytest.fixture
def product_sub_list_union_subscription_1(
    test_product_sub_list_union,
    test_product_type_sub_list_union,
    test_product_block_with_list_union,
    sub_one_subscription_1,
    sub_two_subscription_1,
):
    ProductSubListUnionInactive, _, ProductSubListUnion = test_product_type_sub_list_union
    ProductListUnionBlockForTestInactive, _, _ = test_product_block_with_list_union

    model = ProductSubListUnionInactive.from_product_id(
        product_id=test_product_sub_list_union,
        customer_id=uuid4(),
        status=SubscriptionLifecycle.INITIAL,
        insync=True,
        description="product sub list union sub description",
    )
    model.test_block = ProductListUnionBlockForTestInactive.new(subscription_id=model.subscription_id)
    model.test_block.int_field = 1
    model.test_block.str_field = "blah"
    model.test_block.list_field = [2]
    model.test_block.list_union_blocks = [
        sub_one_subscription_1.test_block,
        sub_two_subscription_1.test_block,
    ]

    model = ProductSubListUnion.from_other_lifecycle(model, status=SubscriptionLifecycle.ACTIVE)
    model.save()

    db.session.commit()

    return model.subscription_id



from sqlalchemy.inspection import inspect

from orchestrator.db import ProcessTable
from orchestrator.db.sorting.sorting import generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

VALID_SORT_KEY_MAP = {
    "created_by": "created_by",
    "started_at": "started",
    "last_status": "status",
    "last_modified_at": "modified",
}
PROCESS_SORT_FUNCTIONS_BY_COLUMN = {
    to_camel(VALID_SORT_KEY_MAP.get(key, key)): generic_column_sort(value)
    for [key, value] in inspect(ProcessTable).columns.items()
}

sort_processes = generic_sort(PROCESS_SORT_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import re
from collections.abc import Generator
from importlib import import_module
from os import listdir, path
from typing import Any, Optional

import structlog

from orchestrator.cli.generator.generator.helpers import get_product_block_file_name, path_to_module, snake_to_camel
from orchestrator.cli.generator.generator.settings import product_generator_settings
from orchestrator.domain.base import ProductBlockModel

logger = structlog.getLogger(__name__)


def get_existing_product_blocks() -> dict[str, Any]:
    def yield_blocks() -> Generator:
        def is_product_block(attribute: Any) -> bool:
            return issubclass(attribute, ProductBlockModel)

        product_blocks_path = product_generator_settings.PRODUCT_BLOCKS_PATH
        product_blocks_module = path_to_module(product_blocks_path)

        if not path.exists(product_blocks_path):
            logger.warning("Product block path does not exist", product_blocks_path=product_blocks_path)
            return

        for pb_file in listdir(product_blocks_path):
            name = pb_file.removesuffix(".py")
            module_name = f"{product_blocks_module}.{name}"

            module = import_module(module_name)

            classes = [obj for _, obj in inspect.getmembers(module, inspect.isclass) if obj.__module__ == module_name]

            for klass in classes:
                if is_product_block(klass):
                    yield klass.__name__, module_name

    return dict(yield_blocks())


def is_constrained_int(field: dict) -> bool:
    return "min_value" in field or "max_value" in field


def is_name_spaced_field_type(field: dict) -> bool:
    return "." in field["type"]


def name_space_get_type(name_spaced_type: str) -> str:
    return name_spaced_type.split(".")[-1]


def get_fields(product_block: dict) -> list[dict]:
    def to_type(field: dict) -> dict:
        if is_constrained_int(field):
            return field | {"type": snake_to_camel(field["name"])}
        elif is_name_spaced_field_type(field):
            return field | {"type": name_space_get_type(field["type"])}
        else:
            return field

    return [to_type(field) for field in product_block["fields"]]


def get_lists_to_generate(fields: list[dict]) -> list[dict]:
    def should_generate(type: str, list_type: Optional[str] = None, **kwargs: Any) -> bool:
        return type == "list" and list_type not in ["str", "int", "bool", "UUID"]

    return [field for field in fields if should_generate(**field)]


def get_name_spaced_types_to_import(fields: list) -> list[tuple]:
    # NOTE: we could make this smarter by grouping imports from the namespace, but isort will handle this for us
    def name_space_split(field: dict) -> tuple[str, str]:
        *namespace, type = field["type"].split(".")
        return ".".join(namespace), type

    return [name_space_split(field) for field in fields if is_name_spaced_field_type(field)]


def get_product_blocks_to_import(lists_to_generate: list, existing_product_blocks: dict) -> list[tuple]:
    return [
        (module, lt["list_type"])
        for lt in lists_to_generate
        if (module := existing_product_blocks.get(f'{lt["list_type"]}Block'))
    ]


def get_product_block_path(product_block: dict) -> str:
    file_name = get_product_block_file_name(product_block)
    return f"{product_generator_settings.PRODUCT_BLOCKS_PATH}/{file_name}.py"


def enrich_product_block(product_block: dict) -> dict:
    def to_block_name() -> str:
        type = product_block["type"]
        name = re.sub("(.)([A-Z][a-z]+)", r"\1 \2", type)
        return re.sub("([a-z0-9])([A-Z])", r"\1 \2", name)

    fields = get_fields(product_block)
    block_name = product_block.get("block_name", to_block_name())

    return product_block | {
        "fields": fields,
        "block_name": block_name,
    }


def generate_product_blocks(context: dict) -> None:
    config = context["config"]
    environment = context["environment"]
    writer = context["writer"]
    python_version = context["python_version"]

    template = environment.get_template("product_block.j2")

    existing_product_blocks = get_existing_product_blocks()

    def generate_product_block(product_block: dict) -> None:
        types_to_import = get_name_spaced_types_to_import(product_block["fields"])

        fields = get_fields(product_block)

        lists_to_generate = get_lists_to_generate(fields)

        product_blocks_to_import = get_product_blocks_to_import(lists_to_generate, existing_product_blocks)

        constrained_ints_to_generate = [field for field in fields if is_constrained_int(field)]

        path = get_product_block_path(product_block)
        content = template.render(
            lists_to_generate=lists_to_generate,
            product_block=enrich_product_block(product_block),
            product_blocks_to_import=product_blocks_to_import,
            constrained_ints_to_generate=constrained_ints_to_generate,
            types_to_import=types_to_import,
            python_version=python_version,
        )

        writer(path, content)

    product_blocks = config.get("product_blocks", [])
    for product_block in product_blocks:
        generate_product_block(product_block)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_non_standard_fixed_inputs(fixed_inputs: list[dict]) -> str:
    return ", ".join(type for fi in fixed_inputs if (type := fi["type"]) not in ["enum", "int", "bool", "str"])


def is_enum(fixed_input: dict) -> bool:
    return fixed_input["type"] == "enum"


def is_enum_of_type(fixed_input: dict, enum_type: str) -> bool:
    return is_enum(fixed_input) and fixed_input["enum_type"] == enum_type


def is_str_enum(fixed_input: dict) -> bool:
    return is_enum_of_type(fixed_input, "str")


def is_int_enum(fixed_input: dict) -> bool:
    return is_enum_of_type(fixed_input, "int")


def convert_enum(fixed_input: dict) -> dict:
    if is_enum(fixed_input):
        return fixed_input | {"type": fixed_input["name"].title()}
    else:
        return fixed_input


def get_str_enum_fixed_inputs(fixed_inputs: list[dict]) -> list[dict]:
    return [convert_enum(fi) for fi in fixed_inputs if is_str_enum(fi)]


def get_int_enum_fixed_inputs(fixed_inputs: list[dict]) -> list[dict]:
    return [convert_enum(fi) for fi in fixed_inputs if is_int_enum(fi)]


def replace_enum_fixed_inputs(fixed_inputs: list[dict]) -> list[dict]:
    return [convert_enum(fi) for fi in fixed_inputs]



from orchestrator.graphql.resolvers.process import resolve_processes
from orchestrator.graphql.resolvers.product import resolve_products
from orchestrator.graphql.resolvers.settings import SettingsMutation, resolve_settings
from orchestrator.graphql.resolvers.subscription import resolve_subscriptions

__all__ = ["resolve_processes", "resolve_products", "resolve_settings", "SettingsMutation", "resolve_subscriptions"]



from typing import Optional

import strawberry

from orchestrator.graphql.schemas.resource_type import ResourceType
from orchestrator.schemas.product_block import ProductBlockSchema


@strawberry.experimental.pydantic.type(model=ProductBlockSchema)
class ProductBlock:
    product_block_id: strawberry.auto
    name: strawberry.auto
    description: strawberry.auto
    tag: strawberry.auto
    status: strawberry.auto
    created_at: strawberry.auto
    end_date: strawberry.auto
    resource_types: Optional[list[ResourceType]]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Union
from uuid import UUID

from more_itertools.more import one
from sqlalchemy.orm import joinedload

from orchestrator.db import ProductTable
from orchestrator.types import UUIDstr


def get_product_by_id(product_id: Union[UUID, UUIDstr]) -> ProductTable:
    """
    Get product by id.

    Args:
        product_id: ProductTable id uuid

    Returns: ProductTable object

    """
    return ProductTable.query.options(joinedload("fixed_inputs")).get(product_id)


def get_product_by_name(name: str) -> ProductTable:
    return ProductTable.query.options(joinedload("fixed_inputs")).filter(ProductTable.name == name).one()


def get_types() -> List[str]:
    return list(
        map(one, ProductTable.query.distinct(ProductTable.product_type).with_entities(ProductTable.product_type))
    )


def get_tags() -> List[str]:
    return list(map(one, ProductTable.query.distinct(ProductTable.tag).with_entities(ProductTable.tag)))
